import contextlib
import itertools
import uuid

import attr
import graham
import marshmallow
import PyQt5.QtCore

import epyqlib.attrsmodel
import epyqlib.treenode
import epyqlib.utils.general
import epyqlib.utils.qt

from natsort import natsorted

# See file COPYING in this source tree
__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'


def create_abbreviation_attribute():
    return attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )


def create_read_only_attribute():
    return attr.ib(
        default=False,
        converter=epyqlib.attrsmodel.two_state_checkbox,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(),
        ),
    )

def sort_multiselect(node):
    if node.visibility is not None:
        node.visibility = natsorted(node.visibility, key=lambda uuid: epyqlib.attrsmodel.name_from_uuid(node, uuid, node.find_root().model))


create_notes_attribute = epyqlib.attrsmodel.create_str_or_none_attribute


@graham.schemify(tag='parameter')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Parameter(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Parameter',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    abbreviation = create_abbreviation_attribute()
    type_name = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    # TODO: CAMPid 1342975467516679768543165421
    default = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Decimal(allow_none=True, as_string=True),
        ),
    )
    minimum = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Decimal(allow_none=True, as_string=True),
        ),
    )
    maximum = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Decimal(allow_none=True, as_string=True),
        ),
    )
    units = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )

    enumeration_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
    )
    epyqlib.attrsmodel.attrib(
        attribute=enumeration_uuid,
        human_name='Enumeration',
        data_display=epyqlib.attrsmodel.name_from_uuid,
        delegate=epyqlib.attrsmodel.RootDelegateCache(
            list_selection_root='enumerations',
        )
    )

    decimal_places = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_int_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(allow_none=True),
        ),
    )
    display_hexadecimal = attr.ib(
        default=False,
        converter=epyqlib.attrsmodel.two_state_checkbox,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(),
        ),
    )
    nv_format = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True)
        ),
    )
    nv_factor = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True)
        ),
    )
    nv_cast = attr.ib(
        default=False,
        converter=epyqlib.attrsmodel.two_state_checkbox,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(),
        ),
    )

    internal_variable = epyqlib.attrsmodel.create_str_or_none_attribute()
    getter_function = epyqlib.attrsmodel.create_str_or_none_attribute()
    setter_function = epyqlib.attrsmodel.create_str_or_none_attribute()
    internal_type = epyqlib.attrsmodel.create_str_or_none_attribute()
    internal_scale_factor = attr.ib(
        default=0,
        converter=int,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(),
        ),
    )

    can_getter = epyqlib.attrsmodel.create_str_or_none_attribute()
    can_setter = epyqlib.attrsmodel.create_str_or_none_attribute()
    sunspec_getter = epyqlib.attrsmodel.create_str_or_none_attribute()
    sunspec_setter = epyqlib.attrsmodel.create_str_or_none_attribute()

    read_only = create_read_only_attribute()

    access_level_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        # converter=lambda x: x if x is None else AccessLevelsAccessLevel(x),
        human_name='Access Level',
        data_display=epyqlib.attrsmodel.name_from_uuid,
        list_selection_root='access level',
    )
    parameter_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
    )
    comment = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    notes = create_notes_attribute()
    original_frame_name = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    original_multiplexer_name = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    original_signal_name = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    visibility = epyqlib.attrsmodel.attr_uuid_list(
        default=None,
        allow_none=True,
    )
    epyqlib.attrsmodel.attrib(
        attribute=visibility,
        human_name='Visibility',
        data_display=epyqlib.attrsmodel.names_from_uuid_list,
        delegate=epyqlib.attrsmodel.RootDelegateCache(
            list_selection_root='visibility',
            multi_select=True,
        )
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return isinstance(node, tuple(self.addable_types().values()))

    @PyQt5.QtCore.pyqtProperty('PyQt_PyObject')
    def pyqtify_minimum(self):
        return epyqlib.utils.qt.pyqtify_get(self, 'minimum')

    @pyqtify_minimum.setter
    def pyqtify_minimum(self, value):
        epyqlib.utils.qt.pyqtify_set(self, 'minimum', value)
        if None not in (value, self.maximum):
            if value > self.maximum:
                self.maximum = value

    @PyQt5.QtCore.pyqtProperty('PyQt_PyObject')
    def pyqtify_maximum(self):
        return epyqlib.utils.qt.pyqtify_get(self, 'maximum')

    @pyqtify_maximum.setter
    def pyqtify_maximum(self, value):
        epyqlib.utils.qt.pyqtify_set(self, 'maximum', value)
        if None not in (value, self.minimum):
            if value < self.minimum:
                self.minimum = value

    def update(self):
        sort_multiselect(self)

    @epyqlib.attrsmodel.check_children
    def check(self, result, models):
        results = []

        print()
        try:
            models.sunspec.root.nodes_by_attribute(
                attribute_value=self.uuid,
                attribute_name='parameter_uuid',
            )
        except epyqlib.treenode.NotFoundError:
            results.append('No linked SunSpec data point found')

        for r in results:
            result.append_child(epyqlib.checkresultmodel.Result(
                node=self,
                message=r,
            ))

        return result

    def uses_interface_item(self):
        return any(v is not None for v in (
            self.internal_variable,
            self.getter_function,
            self.setter_function,
        ))

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move


@graham.schemify(tag='group', register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Group(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Group',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    type_name = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    children = attr.ib(
        default=attr.Factory(list),
        cmp=False,
        repr=False,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(fields=(
                # TODO: would be nice to self reference without a name
                marshmallow.fields.Nested('Group'),
                marshmallow.fields.Nested('Array'),
                marshmallow.fields.Nested('Table'),
                marshmallow.fields.Nested(graham.schema(Parameter)),
            )),
        ),
    )
    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return isinstance(node, tuple(self.addable_types().values()))

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return True

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='enumerations')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Enumerations(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Enumerations Group',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    children = attr.ib(
        default=attr.Factory(list),
        cmp=False,
        repr=False,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(fields=(
                # TODO: would be nice to self reference without a name
                marshmallow.fields.Nested('Enumeration'),
                marshmallow.fields.Nested('AccessLevels'),
            )),
        ),
    )
    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return isinstance(node, tuple(self.addable_types().values()))

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return True

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='array_parameter_element')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@epyqlib.utils.qt.pyqtify_passthrough_properties(
    original='original',
    field_names=(
        'nv_format',
        'nv_factor',
        'nv_cast',
        'access_level_uuid',
        'comment',
        'decimal_places',
        'display_hexadecimal',
        'enumeration_uuid',
        'units',
        'visibility',
    ),
)
@attr.s(hash=False)
class ArrayParameterElement(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Array Parameter Element',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )

    abbreviation = create_abbreviation_attribute()
    notes = create_notes_attribute()
    read_only = create_read_only_attribute()

    # TODO: CAMPid 1342975467516679768543165421
    default = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Decimal(allow_none=True, as_string=True),
        ),
    )
    minimum = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Decimal(allow_none=True, as_string=True),
        ),
    )
    maximum = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Decimal(allow_none=True, as_string=True),
        ),
    )
    nv_format = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True)
        ),
    )
    nv_factor = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True)
        ),
    )
    nv_cast = attr.ib(
        default=False,
        converter=epyqlib.attrsmodel.two_state_checkbox,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(),
        ),
    )
    uuid = epyqlib.attrsmodel.attr_uuid()
    access_level_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        # converter=lambda x: x if x is None else AccessLevelsAccessLevel(x),
        human_name='Access Level',
        data_display=epyqlib.attrsmodel.name_from_uuid,
        list_selection_root='access level',
    )
    enumeration_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
    )
    comment = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    decimal_places = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_int_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(allow_none=True),
        ),
    )
    display_hexadecimal = attr.ib(
        default=False,
        converter=epyqlib.attrsmodel.two_state_checkbox,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(),
        ),
    )
    units = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    visibility = epyqlib.attrsmodel.attr_uuid_list(
        default=None,
        allow_none=True,
    )
    epyqlib.attrsmodel.attrib(
        attribute=visibility,
        human_name='Visibility',
        data_display=epyqlib.attrsmodel.names_from_uuid_list,
        delegate=epyqlib.attrsmodel.RootDelegateCache(
            list_selection_root='visibility',
            multi_select=True,
        )
    )
    original = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=epyqlib.attrsmodel.Reference(),
        ),
    )
    epyqlib.attrsmodel.attrib(
        attribute=original,
        no_column=True,
    )

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    def update(self):
        sort_multiselect(self)

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='array_group_element')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class ArrayGroupElement(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Array Group Element',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    uuid = epyqlib.attrsmodel.attr_uuid()
    original = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=epyqlib.attrsmodel.Reference(),
        ),
    )
    epyqlib.attrsmodel.attrib(
        attribute=original,
        no_column=True,
    )

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


class InvalidArrayLength(Exception):
    pass


@graham.schemify(tag='array', register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Array(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Array',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    length = attr.ib(
        default=1,
        converter=int,
    )
    named_enumerators = attr.ib(
        default=True,
        converter=epyqlib.attrsmodel.two_state_checkbox,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(),
        ),
    )
    children = attr.ib(
        default=attr.Factory(list),
        cmp=False,
        repr=False,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(fields=(
                marshmallow.fields.Nested(graham.schema(Parameter)),
                marshmallow.fields.Nested(graham.schema(ArrayParameterElement)),
                marshmallow.fields.Nested(graham.schema(Group)),
                marshmallow.fields.Nested(graham.schema(ArrayGroupElement)),
            )),
        ),
    )
    uuid = epyqlib.attrsmodel.attr_uuid()

    element_types = {
        Parameter: ArrayParameterElement,
        Group: ArrayGroupElement,
    }

    def __attrs_post_init__(self):
        super().__init__()

        self.length = max(1, len(self.children))

        for child in self.children[1:]:
            if self.children[0].uuid != child.original:
                raise epyqlib.attrsmodel.ConsistencyError(
                    'UUID mismatch: {} != {}'.format(
                        self.children[0].uuid,
                        child.original,
                    )
                )

            child.original = self.children[0]

    @property
    def pyqtify_length(self):
        return epyqlib.utils.qt.pyqtify_get(self, 'length')

    @pyqtify_length.setter
    def pyqtify_length(self, value):
        if value < 1:
            raise InvalidArrayLength('Length must be at least 1')

        if self.children is not None:
            if value < len(self.children):
                for row in range(len(self.children) - 1, value - 1, - 1):
                    self.remove_child(row=row)
            elif 1 <= len(self.children) < value:
                for _ in range(value - len(self.children)):
                    original = self.children[0]
                    type_ = self.element_types[type(original)]
                    self.append_child(type_(original=original))

        epyqlib.utils.qt.pyqtify_set(self, 'length', value)

    @classmethod
    def all_addable_types(cls):
        return epyqlib.attrsmodel.create_addable_types(
            [*cls.element_types.keys(), *cls.element_types.values()],
        )

    def addable_types(self):
        child_types = {type(child) for child in self.children}

        value_types = self.element_types.keys()

        if len(child_types.intersection(set(value_types))) == 0:
            types = value_types
        else:
            # types = (ArrayElement,)
            types = ()

        return epyqlib.attrsmodel.create_addable_types(types)

    def can_drop_on(self, node):
        return isinstance(node, tuple(self.addable_types().values()))

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        if node not in self.children:
            raise epyqlib.attrsmodel.ConsistencyError(
                'Specified node not found in children'
            )

        if len(self.children) > 1:
            return False

        return True

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='table_array_element', register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@epyqlib.utils.qt.pyqtify_passthrough_properties(
    original='original',
    field_names=(
        'name',
        'abbreviation',
        'notes',
        'read_only',
        'access_level_uuid',
        'enumeration_uuid',
        'minimum',
        'maximum',
        'nv_format',
        'nv_factor',
        'nv_cast',
        'comment',
        'units',
        'visibility',
        'display_hexadecimal',
        'default',
        'decimal_places',
    ),
)
@attr.s(hash=False)
class TableArrayElement(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True)
        ),
    )

    abbreviation = create_abbreviation_attribute()
    notes = create_notes_attribute()
    read_only = create_read_only_attribute()

    path = attr.ib(
        factory=tuple,
    )
    epyqlib.attrsmodel.attrib(
        attribute=path,
        no_column=True,
    )
    graham.attrib(
        attribute=path,
        field=graham.fields.Tuple(marshmallow.fields.UUID()),
    )

    access_level_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        # converter=lambda x: x if x is None else AccessLevelsAccessLevel(x),
        human_name='Access Level',
        data_display=epyqlib.attrsmodel.name_from_uuid,
        list_selection_root='access level',
        no_graham=True,
    )

    minimum = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Decimal(allow_none=True, as_string=True),
        ),
    )
    maximum = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Decimal(allow_none=True, as_string=True),
        ),
    )
    nv_format = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True)
        ),
    )
    nv_factor = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True)
        ),
    )
    nv_cast = attr.ib(
        default=False,
        converter=epyqlib.attrsmodel.two_state_checkbox,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(),
        ),
    )
    comment = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    enumeration_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
    )
    units = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    visibility = epyqlib.attrsmodel.attr_uuid_list(
        default=None,
        allow_none=True,
    )
    epyqlib.attrsmodel.attrib(
        attribute=visibility,
        human_name='Visibility',
        data_display=epyqlib.attrsmodel.names_from_uuid_list,
        delegate=epyqlib.attrsmodel.RootDelegateCache(
            list_selection_root='visibility',
            multi_select=True,
        )
    )
    display_hexadecimal = attr.ib(
        default=False,
        converter=epyqlib.attrsmodel.two_state_checkbox,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(),
        ),
    )
    # TODO: CAMPid 1342975467516679768543165421
    default = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Decimal(allow_none=True, as_string=True),
        ),
    )
    decimal_places = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_int_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(allow_none=True),
        ),
    )

    index = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(allow_none=False),
        )
    )
    epyqlib.attrsmodel.attrib(
        attribute=index,
        editable=False,
        no_column=True,
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    original = attr.ib(
        default=None,
        repr=False,
    )
    epyqlib.attrsmodel.attrib(
        attribute=original,
        no_column=True,
    )

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    def update(self):
        sort_multiselect(self)

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='table_group_element', register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@epyqlib.utils.qt.pyqtify_passthrough_properties(
    original='original',
    field_names=('name',),
)
@attr.s(hash=False)
class TableGroupElement(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True)
        ),
    )

    path = attr.ib(
        factory=tuple,
    )
    epyqlib.attrsmodel.attrib(
        attribute=path,
        no_column=True,
    )
    graham.attrib(
        attribute=path,
        field=graham.fields.Tuple(marshmallow.fields.UUID()),
    )

    children = attr.ib(
        default=attr.Factory(list),
        cmp=False,
        repr=False,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(fields=(
                marshmallow.fields.Nested('TableGroupElement'),
                marshmallow.fields.Nested(graham.schema(TableArrayElement)),
            )),
        ),
    )

    axis = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        )
    )
    epyqlib.attrsmodel.attrib(
        attribute=axis,
        editable=False,
        no_column=True,
    )

    curve_index = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(allow_none=True),
        )
    )
    epyqlib.attrsmodel.attrib(
        attribute=curve_index,
        editable=False,
        no_column=True,
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    original = attr.ib(default=None)
    epyqlib.attrsmodel.attrib(
        attribute=original,
        no_column=True,
    )

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    def can_delete(self, node=None):
        return False

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='table_enumeration_reference')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class TableEnumerationReference(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Enumeration Reference',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    enumeration_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        human_name='Enumeration',
        data_display=epyqlib.attrsmodel.name_from_uuid,
        list_selection_root='enumerations',
    )

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    can_delete = epyqlib.attrsmodel.childless_can_delete

    def link(self, enumeration):
        self.enumeration_uuid = enumeration.uuid

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='table', register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Table(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Table',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )

    can_getter = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )

    can_setter = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )

    active_curve_getter = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )

    active_curve_setter = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )

    children = attr.ib(
        default=attr.Factory(list),
        cmp=False,
        repr=False,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(
                    marshmallow.fields.Nested(graham.schema(Array)),
                    marshmallow.fields.Nested(graham.schema(Group)),
                    marshmallow.fields.Nested(graham.schema(
                        TableEnumerationReference,
                    )),
                    marshmallow.fields.Nested(graham.schema(TableGroupElement)),
                ),
            ),
        ),
    )
    uuid = epyqlib.attrsmodel.attr_uuid()

    group = attr.ib(default=None)
    epyqlib.attrsmodel.attrib(
        attribute=group,
        no_column=True,
    )

    combinations = attr.ib(default=None)
    epyqlib.attrsmodel.attrib(
        attribute=combinations,
        no_column=True,
    )

    arrays = attr.ib(default=None)
    epyqlib.attrsmodel.attrib(
        attribute=arrays,
        no_column=True,
    )

    arrays_and_groups = attr.ib(default=None)
    epyqlib.attrsmodel.attrib(
        attribute=arrays_and_groups,
        no_column=True,
    )

    curve_group_combinations = attr.ib(default=None)
    epyqlib.attrsmodel.attrib(
        attribute=curve_group_combinations,
        no_column=True,
    )

    def __attrs_post_init__(self):
        super().__init__()

        self._monitor_children = True
        self.pyqt_signals.child_added_complete.connect(self.update)
        self.pyqt_signals.child_removed_complete.connect(self.update)

        self.pyqt_signals.child_added_complete.connect(self.array_added)
        self.pyqt_signals.child_removed_complete.connect(self.array_removed)

        self.array_connections = {}

    @contextlib.contextmanager
    def _ignore_children(self):
        self._monitor_children = False
        yield
        self._monitor_children = True

    def array_added(self, array):
        if not isinstance(array, Array):
            return

        self.array_connections[array] = epyqlib.utils.qt.Connections(
            signal=epyqlib.utils.qt.pyqtify_signals(array).length,
            slot=self.update,
        )


    def array_removed(self, array):
        if not isinstance(array, Array):
            return

        connections = self.array_connections.pop(array)
        connections.disconnect()

    def update_array_connections(self):
        for array in list(self.array_connections.keys()):
            self.array_removed(array)

        arrays = [
            child
            for child in self.children
            if isinstance(child, Array)
        ]

        for array in arrays:
            self.array_added(array)

    def update(self, changed=None):
        if not self._monitor_children:
            return

        self.update_array_connections()

        old_groups = [
            child
            for child in self.children
            if isinstance(child, TableGroupElement)
        ]

        if len(old_groups) == 1:
            old_group, = old_groups
        elif len(old_groups) < 1:
            old_group = None
        else:
            raise Exception('Too many old groups found while updating ')

        root = self.find_root()

        enumerations = []

        for child in self.children:
            if not isinstance(child, TableEnumerationReference):
                continue

            if child.enumeration_uuid is None:
                continue

            enumeration = root.model.node_from_uuid(child.enumeration_uuid)

            enumerations.append(enumeration.children)

        arrays = [
            child
            for child in self.children
            if isinstance(child, (Array, Group))
        ]
        self.arrays_and_groups = arrays
        self.arrays = [
            child
            for child in self.children
            if isinstance(child, Array)
        ]
        self.groups = [
            child
            for child in self.children
            if isinstance(child, Group)
        ]

        with self._ignore_children():
            if old_group is None:
                old_group = TableGroupElement(
                    name='Tree',
                )
                self.append_child(old_group)

            self.group = old_group

            nodes = old_group.recursively_remove_children()

            old_by_path = {
                node.path: node
                for node in nodes
            }

        product = list(itertools.product(*enumerations))

        self.combinations = product

        self.curve_group_combinations = tuple(
            epyqlib.utils.general.ordered_unique(
                tuple(
                    x
                    for x in itertools.takewhile(
                        lambda y: y.tree_parent.name != 'Curves',
                        combination,
                    )
                )
                for combination in self.combinations
            )
        )

        model = self.find_root().model

        for combination in product:
            present = old_group

            path = ()

            for layer in combination:
                path += (layer.uuid,)

                previous = old_by_path.get(path)
                if previous is None:
                    current = TableGroupElement(
                        original=layer,
                        path=path,
                    )
                    old_by_path[path] = current
                else:
                    current = previous
                    if current.original is None:
                        current.original = current.path[-1]
                    if isinstance(current.original, uuid.UUID):
                        current.original = model.node_from_uuid(current.original)

                if layer.tree_parent.name == 'Curves':
                    current.curve_index = int(layer.value)

                if current.tree_parent is None:
                    present.append_child(current)

                present = current

            axes = ['x', 'y', 'z']
            axes_iterator = iter(axes)
            for array in arrays:
                if isinstance(array, Array):
                    try:
                        axis = next(axes_iterator)
                    except StopIteration:
                        raise
                else:
                    axis = None
                array_path = path + (array.uuid,)
                previous = old_by_path.get(array_path)
                if previous is None:
                    current = TableGroupElement(
                        original=array,
                        path=array_path,
                    )
                    old_by_path[array_path] = current
                else:
                    current = previous
                    if current.original is None:
                        current.original = current.path[-1]
                    if isinstance(current.original, uuid.UUID):
                        current.original = model.node_from_uuid(
                            current.original
                        )

                current.axis = axis

                if current.tree_parent is None:
                    present.append_child(current)

                for index, element in enumerate(array.children):
                    element_path = array_path + (element.uuid,)
                    previous_element = old_by_path.get(element_path)
                    if previous_element is None:
                        current_element = TableArrayElement(
                            original=element,
                            path=element_path,
                        )
                        old_by_path[element_path] = current_element
                    else:
                        current_element = previous_element
                        if current_element.original is None:
                            current_element.original = current_element.path[-1]
                        if isinstance(current_element.original, uuid.UUID):
                            current_element.original = model.node_from_uuid(
                                current_element.original
                            )

                    current_element.index = index

                    if current_element.tree_parent is None:
                        current.append_child(current_element)

    def addable_types(self):
        return epyqlib.attrsmodel.create_addable_types((
            TableEnumerationReference,
            Array,
            Group,
        ))

    @classmethod
    def all_addable_types(cls):
        return epyqlib.attrsmodel.create_addable_types((
            TableEnumerationReference,
            Array,
            Group,
            TableGroupElement,
        ))

    def can_drop_on(self, node):
        return isinstance(node, tuple(self.addable_types().values()))

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        if node not in self.children:
            raise epyqlib.attrsmodel.ConsistencyError(
                'Specified node not found in children'
            )

        return not isinstance(node, TableGroupElement)

    def internal_move(self, node, node_to_insert_before):
        with self._ignore_children():
            self.remove_child(child=node)
            if node_to_insert_before is None:
                self.append_child(node)
            else:
                self.insert_child(
                    i=self.children.index(node_to_insert_before),
                    child=node,
                )

        return True

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='enumerator')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Enumerator(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Enumerator',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    value = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(allow_none=True),
        ),
    )
    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='sunspec_enumerator')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class SunSpecEnumerator(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Sunspec Enumerator',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    label = attr.ib(
        default='',
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    description = attr.ib(
        default='',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    notes = create_notes_attribute()
    value = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(allow_none=True),
        ),
    )
    type = attr.ib(
        default='',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='enumeration', register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Enumeration(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Enumeration',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    children = attr.ib(
        default=attr.Factory(list),
        repr=False,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(fields=(
                marshmallow.fields.Nested(graham.schema(Enumerator)),
                marshmallow.fields.Nested(graham.schema(SunSpecEnumerator)),
            )),
        ),
    )
    # children = attr.ib(
    #     default=attr.Factory(list),
    #     metadata=graham.create_metadata(
    #         field=marshmallow.fields.List(
    #             marshmallow.fields.Nested(graham.schema(Enumerator)),
    #         ),
    #     ),
    # )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def items(self):
        for child in self.children:
            yield (child.name, child.value)

    def values(self):
        for child in self.children:
            yield child.value

    def can_drop_on(self, node):
        return isinstance(node, Enumerator)

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return True

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='access_level')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class AccessLevel(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Access Level',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    value = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(allow_none=True),
        ),
    )
    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag='access_levels', register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class AccessLevels(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Access Levels',
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    children = attr.ib(
        default=attr.Factory(list),
        repr=False,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(fields=(
                marshmallow.fields.Nested(graham.schema(AccessLevel)),
            )),
        ),
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def items(self):
        for child in self.children:
            yield (child.name, child.value)

    def values(self):
        for child in self.children:
            yield child.value

    def can_drop_on(self, node):
        return False

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return True

    def by_name(self, name):
        level, = (
            level
            for level in self.children
            if level.name.casefold() == name.casefold()
        )

        return level

    def default(self):
        return min(self.children, key=lambda x: x.value)

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


Root = epyqlib.attrsmodel.Root(
    default_name='Parameters',
    valid_types=(Parameter, Group, Enumerations)
)

types = epyqlib.attrsmodel.Types(
    types=(
        Root,
        Parameter,
        Group,
        Array,
        ArrayGroupElement,
        ArrayParameterElement,
        Enumeration,
        Enumerator,
        SunSpecEnumerator,
        Enumerations,
        AccessLevel,
        AccessLevels,
        Table,
        TableEnumerationReference,
        TableArrayElement,
        TableGroupElement,
    ),
    external_list_selection_roots={'sunspec types'},
)


# TODO: CAMPid 943896754217967154269254167
def merge(name, *types):
    return tuple((x, name) for x in types)


columns = epyqlib.attrsmodel.columns(
    merge('name', *types.types.values()),
    merge('abbreviation', Parameter, ArrayParameterElement, TableArrayElement),
    (
        merge('type_name', Parameter, Group)
        + merge('type', SunSpecEnumerator)
    ),
    merge('length', Array),
    merge('named_enumerators', Array),
    merge(
        'units',
        Parameter,
        ArrayParameterElement,
        TableArrayElement,
    ),
    merge(
        'enumeration_uuid',
        Parameter,
        TableEnumerationReference,
        ArrayParameterElement,
        TableArrayElement,
    ),

    merge('value', Enumerator, SunSpecEnumerator, AccessLevel),
    merge(
        'default',
        Parameter,
        ArrayParameterElement,
        TableArrayElement,
    ),
    merge(
        'minimum',
        Parameter,
        ArrayParameterElement,
        TableArrayElement,
    ),
    merge(
        'maximum',
        Parameter,
        ArrayParameterElement,
        TableArrayElement,
    ),

    merge('label', SunSpecEnumerator),

    merge('internal_variable', Parameter),
    merge('getter_function', Parameter),
    merge('setter_function', Parameter),
    merge('internal_type', Parameter),
    merge('internal_scale_factor', Parameter),

    merge('can_getter', Table, Parameter),
    merge('can_setter', Table, Parameter),
    merge('sunspec_getter', Parameter),
    merge('sunspec_setter', Parameter),
    merge('active_curve_getter', Table),
    merge('active_curve_setter', Table),
    merge(
        'nv_format',
        Parameter,
        ArrayParameterElement,
        TableArrayElement,
    ),
    merge(
        'nv_factor',
        Parameter,
        TableArrayElement,
        ArrayParameterElement,
    ),
    merge(
        'nv_cast',
        Parameter,
        TableArrayElement,
        ArrayParameterElement,
    ),
    merge('read_only', Parameter, ArrayParameterElement, TableArrayElement),
    merge(
        'access_level_uuid',
        Parameter,
        TableArrayElement,
        ArrayParameterElement,
    ),
    merge(
        'visibility',
        Parameter,
        ArrayParameterElement,
        TableArrayElement,
    ),

    merge(
        'display_hexadecimal',
        Parameter,
        ArrayParameterElement,
        TableArrayElement,
    ),
    merge(
        'decimal_places',
        Parameter,
        ArrayParameterElement,
        TableArrayElement,
    ),

    (
        merge(
            'comment',
            Parameter,
            ArrayParameterElement,
            TableArrayElement,
        )
        +
        merge('description', SunSpecEnumerator)
    ),

    merge(
        'notes',
        Parameter,
        SunSpecEnumerator,
        ArrayParameterElement,
        TableArrayElement,
    ),

    merge('original_frame_name', Parameter),
    merge('original_multiplexer_name', Parameter),
    merge('original_signal_name', Parameter),

    merge('parameter_uuid', Parameter),
    merge('uuid', *types.types.values()),

)



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import logging
logger = logging.getLogger(__name__)

import attr
import collections
import can
import contextlib
import decimal
import enum
from epyqlib.abstractcolumns import AbstractColumns
import epyqlib.attrsmodel
import epyqlib.canneo
import epyqlib.pm.valuesetmodel
import epyqlib.twisted.busproxy
import epyqlib.twisted.nvs
import epyqlib.utils.general
import epyqlib.utils.qt
import epyqlib.utils.twisted
import functools
import itertools
import json
import epyqlib.pyqabstractitemmodel
from epyqlib.treenode import TreeNode
from PyQt5.QtCore import (Qt, QVariant, QModelIndex, pyqtSignal, pyqtSlot)
from PyQt5 import QtGui, QtWidgets
from PyQt5.QtWidgets import QFileDialog
import qtawesome
import re
import textwrap
import time
import twisted.internet.defer
import twisted.internet.task
import uuid
import weakref

# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class Columns(AbstractColumns):
    _members = [
        'name',
        'read_only',
        'factory',
        'value',
        'saturate',
        'reset',
        'clear',
        'scratch',
        'user_default',
        'factory_default',
        'minimum',
        'maximum',
        'comment',
    ]

Columns.indexes = Columns.indexes()


diffable_columns = [
    Columns.indexes.value,
    Columns.indexes.scratch,
    Columns.indexes.user_default,
    Columns.indexes.factory_default,
    Columns.indexes.minimum,
    Columns.indexes.maximum,
]

diff_highlight = QtGui.QColor('orange')
diff_highlight.setAlphaF(0.4)

reference_highlight = QtGui.QColor('green')
reference_highlight.setAlphaF(0.2)


class NoNv(Exception):
    pass


class NotFoundError(Exception):
    pass


class TooManyMetaSignalsError(Exception):
    pass


class SimultaneousTransactionsError(Exception):
    pass


@attr.s
class Configuration:
    set_frame = attr.ib()
    status_frame = attr.ib()
    to_nv_command = attr.ib()
    to_nv_status = attr.ib()
    read_write_signal = attr.ib()
    read_write_status_signal = attr.ib()
    meta_signal = attr.ib()
    nv_save_in_progress_name = attr.ib()

configurations = {
    'original': Configuration(
        set_frame='CommandSetNVParam',
        status_frame='StatusNVParam',
        to_nv_command='SaveToEE_command',
        to_nv_status='SaveToEE_status',
        read_write_signal='ReadParam_command',
        read_write_status_signal='ReadParam_status',
        meta_signal=None,
        nv_save_in_progress_name = None,
    ),
    'j1939': Configuration(
        set_frame='ParameterQuery',
        status_frame='ParameterResponse',
        to_nv_command='SaveToEE_command',
        to_nv_status='SaveToEE_status',
        read_write_signal='ReadParam_command',
        read_write_status_signal = 'ReadParam_status',
        meta_signal='Meta',
        nv_save_in_progress_name = 'eeSaveInProgress',
    )
}


@attr.s(hash=True)
class Group(TreeNode):
    fields = attr.ib(default=attr.Factory(Columns))

    def __attrs_post_init__(self):
        super().__init__()


class MetaEnum(epyqlib.utils.general.AutoNumberIntEnum):
    value = 0
    user_default = 1
    factory_default = 2
    minimum = 3
    maximum = 4


meta_limits = (MetaEnum.minimum, MetaEnum.maximum)
meta_limits_first = (
    meta_limits + tuple(sorted(set(MetaEnum) - set(meta_limits)))
)

MetaEnum.non_value = tuple(sorted(set(MetaEnum) - {MetaEnum.value}))

meta_column_indexes = tuple(
    getattr(Columns.indexes, meta.name)
    for meta in MetaEnum
)


dynamic_column_indexes_by_meta = {
    MetaEnum.value: [
        Columns.indexes.value,
        Columns.indexes.saturate,
        Columns.indexes.reset,
        Columns.indexes.clear,
    ],
    **{
        meta: [getattr(Columns.indexes, meta.name)]
        for meta in MetaEnum.non_value
    },
}


column_index_by_meta = {
    meta: getattr(Columns.indexes, meta.name)
    for meta in MetaEnum
}


@attr.s
@epyqlib.utils.general.enumerated_attrs(MetaEnum, default=None)
class Meta:
    pass


class Nvs(TreeNode, epyqlib.canneo.QtCanListener):
    changed = epyqlib.utils.qt.Signal(
        TreeNode,
        int,
        TreeNode,
        int,
        list,
    )
    activity_started = epyqlib.utils.qt.Signal(str)
    activity_ended = epyqlib.utils.qt.Signal(str)

    def __init__(self, neo, bus=None, stop_cyclic=None, start_cyclic=None,
                 configuration=None, hierarchy=None, metas=(MetaEnum.value,),
                 access_level_path=None, access_password_path=None,
                 parent=None):
        TreeNode.__init__(self)
        epyqlib.canneo.QtCanListener.__init__(self, parent=parent)

        if configuration is None:
            configuration = 'original'

        self.stop_cyclic = stop_cyclic
        self.start_cyclic = start_cyclic
        self.configuration = configurations[configuration]

        from twisted.internet import reactor
        self.protocol = epyqlib.twisted.nvs.Protocol()
        self.transport = epyqlib.twisted.busproxy.BusProxy(
            protocol=self.protocol,
            reactor=reactor,
            bus=bus)

        self.bus = bus
        self.neo = neo
        self.message_received_signal.connect(self.message_received)

        self.access_level_node = None
        if access_level_path is not None:
            self.access_level_node = self.neo.signal_by_path(*access_level_path)
            self.access_level_node.write_only = True

        self.password_node = None
        if access_password_path is not None:
            self.password_node = self.neo.signal_by_path(*access_password_path)

        self.set_frames = [f for f in self.neo.frames
                       if f.name == self.configuration.set_frame]
        try:
            self.set_frames = self.set_frames[0]
        except IndexError:
            # TODO: custom error
            raise NoNv()

        self.set_frames = self.set_frames.multiplex_frames
        self.status_frames = [
            f for f in self.neo.frames
            if f.name == self.configuration.status_frame
        ][0].multiplex_frames

        self.save_frame = None
        self.save_signal = None
        self.save_value = None
        self.nv_save_in_progress_signal = None
        self.confirm_save_frame = None
        self.confirm_save_multiplex_value = None
        self.confirm_save_signal = None
        self.confirm_save_value = None
        for frame in self.set_frames.values():
            for signal in frame.signals:
                if signal.name == self.configuration.to_nv_command:
                    for key, value in signal.enumeration.items():
                        if value == 'Enable':
                            self.save_frame = frame
                            self.save_signal = signal
                            self.save_value = key

        save_status_name = self.configuration.to_nv_status
        for frame in self.status_frames.values():
            for signal in frame.signals:
                if signal.name == save_status_name:
                    for key, value in signal.enumeration.items():
                        if value == 'Enable':
                            self.confirm_save_frame = frame
                            self.confirm_save_multiplex_value = signal.multiplex
                            self.confirm_save_signal = signal
                            self.confirm_save_value = key

        for frame in self.set_frames.values():
            for signal in frame.signals:
                if signal.name == self.configuration.nv_save_in_progress_name:
                    self.nv_save_in_progress_signal = signal

        if self.confirm_save_frame is None:
            raise Exception(
                "'{}' signal not found in NV parameter interface".format(
                    save_status_name
                ))

        self.nv_by_path = {}
        # TODO: kind of an ugly manual way to connect this
        self.status_frames[0].set_frame = self.set_frames[0]
        for value, frame in self.set_frames.items():
            signals = [s for s in frame.signals]
            signals = [s for s in signals if s.multiplex is not 'Multiplexor']
            signals = [
                s for s in signals
                if s.name not in [
                    self.configuration.read_write_signal,
                    self.configuration.meta_signal,
                    '{}_MUX'.format(self.configuration.set_frame)
                ]
            ]

            if len(signals) > 0:
                def ignore_timeout(failure):
                    acceptable_errors = (
                        epyqlib.twisted.nvs.RequestTimeoutError,
                        epyqlib.twisted.nvs.SendFailedError,
                        epyqlib.twisted.nvs.CanceledError,
                    )
                    if failure.type in acceptable_errors:
                        return None

                    return epyqlib.utils.twisted.errbackhook(
                        failure)

                def send(signals=None, all_signals=signals):
                    if signals is None:
                        signals = all_signals

                    d = twisted.internet.defer.Deferred()
                    d.callback(None)

                    for enumerator in metas:
                        d.addCallback(lambda _: self.protocol.write_multiple(
                            nv_signals=signals,
                            meta=enumerator,
                            priority=epyqlib.twisted.nvs.Priority.user
                        ))
                    d.addErrback(ignore_timeout)

                frame._send.connect(send)

            frame.parameter_signals = []
            for nv in signals:
                if nv.name not in [self.configuration.to_nv_command]:
                    self.nv_by_path[nv.signal_path()] = nv
                    frame.parameter_signals.append(nv)
                    nv.changed.connect(self.changed)

                try:
                    nv.frame.status_frame = self.status_frames[value]
                except KeyError:
                    raise Exception('Unable to find: {}:{} with mux id {}'.format(
                        self.configuration.status_frame,
                        frame.mux_name,
                        frame.mux_value,
                    ))

                self.status_frames[value].set_frame = nv.frame

                search = (s for s in self.status_frames[value].signals
                          if s.start_bit == nv.start_bit and s.signal_size == nv.signal_size)
                try:
                    nv.status_signal, = search
                except ValueError:
                    raise Exception(
                        'NV status signal not found for {}:{}'.format(nv.frame.mux_name, nv.name)
                    )
                nv.status_signal.set_signal = nv


        unreferenced_paths = set(self.nv_by_path)
        if hierarchy is not None:
            def handle(children, tree_parent,
                       unreferenced_paths=unreferenced_paths):
                unreferenced_groups = []
                logger.info(children, tree_parent)
                for child in children:
                    logger.info(child)
                    if isinstance(child, dict):
                        group = Group(
                            fields=Columns(
                                name=child['name'],
                            )
                        )
                        tree_parent.append_child(group)
                        logger.info('added group: {}'.format(group.fields.name))
                        if child.get('unreferenced'):
                            unreferenced_groups.append(group)
                        else:

                            unreferenced_groups.extend(handle(
                                children=child.get('children', ()),
                                tree_parent=group
                            ))
                    else:
                        path = ('ParameterQuery',) + tuple(child)
                        if path in unreferenced_paths:
                            tree_parent.append_child(self.nv_by_path[path])
                            unreferenced_paths.discard(path)
                        elif path not in self.nv_by_path:
                            logger.info('Unknown parameter referenced: {}'
                                  .format(path))
                        else:
                            raise Exception('Attempted to put parameter in '
                                            'multiple groups: {}'.format(path))

                return tuple(g for g in unreferenced_groups if g is not None)

            unreferenced_groups = handle(children=hierarchy['children'],
                                      tree_parent=self)

            logger.info('\\/ \\/ \\/ unreferenced parameter paths')
            logger.info(
                json.dumps(
                    tuple(p[1:] for p in sorted(unreferenced_paths)),
                    indent=4
                )
            )
            logger.info('/\\ /\\ /\\ unreferenced parameter paths')
        else:
            unreferenced_groups = (self,)

        for group in unreferenced_groups:
            for path in unreferenced_paths:
                group.append_child(self.nv_by_path[path])

        def remove_empty_groups(node):
            if isinstance(node, Group) and len(node.children) == 0:
                node.tree_parent.remove_child(child=node)
            else:
                for child in node.children:
                    remove_empty_groups(child)

        remove_empty_groups(self)

        duplicate_names = set()
        found_names = set()
        for child in self.all_nv():
            name = child.fields.name
            if name not in found_names:
                found_names.add(name)
            else:
                duplicate_names.add(name)

        if len(duplicate_names) > 0:
            raise Exception('Duplicate NV parameter names found: {}'.format(
                ', '.join(duplicate_names)))

        missing_read_write_signals = []

        for i, set_frame in self.set_frames.items():
            status_frame = self.status_frames[i]
            try:
                set_frame.command_signal, = (
                    s
                    for s in set_frame.signals
                    if s.name == self.configuration.read_write_signal
                )
            except ValueError:
                missing_read_write_signals.append('{}:{}:{}'.format(
                    set_frame.name,
                    set_frame.mux_name,
                    self.configuration.read_write_signal,
                ))

            try:
                status_frame.command_signal, = (
                    s
                    for s in status_frame.signals
                    if s.name == self.configuration.read_write_status_signal
                )
            except ValueError:
                missing_read_write_signals.append('{}:{}:{}'.format(
                    status_frame.name,
                    status_frame.mux_name,
                    self.configuration.read_write_status_signal,
                ))

        if len(missing_read_write_signals) > 0:
            raise Exception('Unable to find needed signals: {}'.format(
                ', '.join(missing_read_write_signals),
            ))

        self.cyclic_reader = CyclicReader(
            nvs=self.all_nv(),
            read_call=self.read_all_from_device,
            metas=metas,
        )

    def terminate(self):
        self.cancel_cyclic_read_all()

    def set_stale(self):
        for nv in self.all_nv():
            nv.set_stale()

    def set_bus(self, bus):
        self.transport.set_bus(bus=bus)
        self.bus = bus

    def all_nv(self):
        def visit(node, all):
            if isinstance(node, Nv):
                all.add(node)
            else:
                for child in node.children:
                    visit(child, all)

        all = set()
        visit(self, all)

        return all

    def names(self):
        return '\n'.join([n.fields.name for n in self.all_nv()])

    def write_all_to_device(
            self,
            only_these=None,
            values=(),
            callback=None,
            meta=None,
            background=False,
    ):
        return self._read_write_all(
            read=False,
            only_these=only_these,
            values=values,
            callback=callback,
            meta=meta,
            background=background,
        )

    def read_all_from_device(
            self,
            only_these=None,
            callback=None,
            meta=None,
            background=False,
    ):
        return self._read_write_all(
            read=True,
            only_these=only_these,
            callback=callback,
            meta=meta,
            background=background,
        )

    def _read_write_all(
            self,
            read,
            only_these=None,
            values=(),
            callback=None,
            meta=None,
            background=False,
    ):
        if meta is None:
            meta = meta_limits_first

        activity = ('Reading from device' if read
                    else 'Writing to device')

        if not background:
            self.activity_started.emit('{}...'.format(activity))
        d = twisted.internet.defer.Deferred()
        d.callback(None)

        def handle_frame(frame, signals, enumerator):
            if read:
                d.addCallback(
                    lambda _, enumerator=enumerator: self.protocol.read_multiple(
                        nv_signals=signals,
                        meta=enumerator,
                        priority=epyqlib.twisted.nvs.Priority.user,
                        passive=True,
                        all_values=True,
                    )
                )
            elif frame.read_write.min <= 0:
                not_none_signals = {}
                for signal in signals:
                    values_key = (signal, enumerator)
                    if values_key in values:
                        value = values[values_key]
                    elif enumerator == MetaEnum.value:
                        value = signal.value
                    else:
                        value = getattr(signal.meta, enumerator.name).value

                    if value is not None:
                        not_none_signals[signal] = value

                if len(not_none_signals) == 0:
                    return

                d.addCallback(
                    lambda _, enumerator=enumerator, not_none_signals=not_none_signals: self.protocol.write_multiple(
                        nv_signals=not_none_signals,
                        meta=enumerator,
                        priority=epyqlib.twisted.nvs.Priority.user,
                        passive=True,
                        all_values=True,
                    )
                )
            else:
                return

            if callback is not None:
                d.addCallback(callback)

        if only_these is None:
            only_these = self.all_nv()

        frames = set(nv.frame for nv in only_these)
        for frame in frames:
            signals = tuple(nv for nv in only_these
                            if nv.frame is frame)

            frame.update_from_signals()
            for enumerator in meta:
                handle_frame(
                    frame=frame,
                    signals=signals,
                    enumerator=enumerator,
                )

        if not background:
            d.addCallback(epyqlib.utils.twisted.detour_result,
                          self.activity_ended.emit,
                          'Finished {}...'.format(activity.lower()))
            d.addErrback(epyqlib.utils.twisted.detour_result,
                         self.activity_ended.emit,
                         'Failed while {}...'.format(activity.lower()))

        return d

    def message_received(self, msg):
        if (msg.arbitration_id == self.status_frames[0].id
                and msg.id_type == self.status_frames[0].extended):
            multiplex_message, multiplex_value =\
                self.neo.get_multiplex(msg)

            if multiplex_message is None:
                return

            if multiplex_value is not None and multiplex_message in self.status_frames.values():
                values = multiplex_message.unpack(msg.data, only_return=True)

                if multiplex_message.meta_signal is not None:
                    meta = epyqlib.nv.MetaEnum(
                        values[multiplex_message.meta_signal],
                    )
                    if meta != epyqlib.nv.MetaEnum.value:
                        return

                multiplex_message.unpack(msg.data)
                # multiplex_message.frame.update_canneo_from_matrix_signals()

                status_signals = multiplex_message.signals
                sort_key = lambda s: s.start_bit
                status_signals = sorted(status_signals, key=sort_key)
                set_signals = multiplex_message.set_frame.signals
                set_signals = sorted(set_signals, key=sort_key)
                for status, set in zip(status_signals, set_signals):
                    set.set_value(status.value)

    def unique(self):
        # TODO: actually identify the object
        return '-'

    def to_dict(self, include_secrets=False):
        d = {}
        for child in self.all_nv():
            if include_secrets or not child.secret:
                d[child.fields.name] = child.scratch.get_human_value(
                    for_file=True,
                )

        return d

    def to_value_set(self, include_secrets=False):
        value_set = epyqlib.pm.valuesetmodel.create_blank()

        for child in self.all_nv():
            if not child.secret:
                parameter = epyqlib.pm.valuesetmodel.Parameter(
                    name=child.fields.name,
                    value=child.get_human_value(for_file=True),
                    user_default=child.meta.user_default.get_human_value(
                        for_file=True
                    ),
                    factory_default=child.meta.factory_default.get_human_value(
                        for_file=True
                    ),
                    minimum=child.meta.minimum.get_human_value(for_file=True),
                    maximum=child.meta.maximum.get_human_value(for_file=True),
                    readable=not child.is_write_only(),
                    writable=not child.is_read_only(),
                    parameter_uuid=child.parameter_uuid,
                )
            elif include_secrets:
                def format_float(value):
                    was_secret = child.secret
                    child.secret = False
                    result = child.format_float(value=value, for_file=True)
                    child.secret = was_secret

                    return result

                parameter = epyqlib.pm.valuesetmodel.Parameter(
                    name=child.fields.name,
                    value=format_float(value=0),
                    user_default=format_float(value=0),
                    factory_default=format_float(value=0),
                    minimum=format_float(value=child.min),
                    maximum=format_float(value=child.max),
                    parameter_uuid=child.parameter_uuid,
                )
            else:
                continue

            value_set.model.root.append_child(parameter)

        return value_set

    def from_dict(self, d):
        only_in_file = list(d.keys())

        for child in self.all_nv():
            value = d.get(child.fields.name, None)
            if value is not None:
                child.scratch.set_human_value(decimal.Decimal(value))
                child.fields.scratch = child.scratch.full_string
                only_in_file.remove(child.fields.name)
            else:
                logger.info("Nv value named '{}' not found when loading from dict"
                      .format(child.fields.name))

        for name in only_in_file:
            logger.info("Unrecognized NV value named '{}' found when loading "
                  "from dict".format(name))

    def from_value_set(self, value_set):
        parameter_nodes = value_set.model.root.nodes_by_filter(
            filter=lambda node: isinstance(
                node,
                epyqlib.pm.valuesetmodel.Parameter,
            ),
        )
        only_in_file = {
            (parameter, meta)
            for parameter, meta in itertools.product(
                parameter_nodes,
                epyqlib.nv.MetaEnum,
            )
        }

        d = collections.defaultdict(list)
        for parameter in parameter_nodes:
            d[parameter.name].append(parameter)
            d[parameter.parameter_uuid].append(parameter)

        for child in self.all_nv():
            name = child.fields.name

            parameters = d.get(child.parameter_uuid, [])
            if len(parameters) == 0:
                parameters = d.get(name, [])

            not_found_format = (
                "Nv value named '{}' ({{}}) not found when loading "
                "from value set".format(name)
            )

            if len(parameters) == 1:
                parameter, = parameters

                for meta in MetaEnum:
                    only_in_file.discard((name, meta))

                if parameter.value is not None:
                    child.set_human_value(parameter.value)
                else:
                    logger.info(not_found_format.format('value'))

                for meta in MetaEnum:
                    if meta == MetaEnum.value:
                        continue

                    v = getattr(parameter, meta.name)
                    if v is not None:
                        child.set_meta(
                            data=v,
                            meta=meta,
                            check_range=False,
                        )
                    else:
                        logger.info(not_found_format.format(meta.name))
            elif len(parameters) > 1:
                logger.info(
                    "Nv value named '{}' occurred {} times when loading "
                    "from value set".format(name, len(parameters)),
                )
            else:
                logger.info(
                    "Nv value named '{}' not found when loading from "
                    "value set".format(name),
                )

        for name, meta in sorted(only_in_file):
            logger.info("Unrecognized NV value named '{}' ({}) found when loading "
                  "from value set".format(name, meta.name))

    def defaults_from_dict(
            self,
            d,
            default_metas=(MetaEnum.factory_default,),
    ):
        only_in_file = list(d.keys())

        for child in self.all_nv():
            value = d.get(child.fields.name, None)
            if value is not None:
                for meta in default_metas:
                    child.set_meta(
                        data=value,
                        meta=meta,
                    )
                    self.changed.emit(
                        child, meta_column_indexes[meta],
                        child, meta_column_indexes[meta],
                        [Qt.DisplayRole],
                    )
                only_in_file.remove(child.fields.name)
            else:
                logger.info("Nv value named '{}' not found when loading from dict"
                      .format(child.fields.name))

        for name in only_in_file:
            logger.info("Unrecognized NV value named '{}' found when loading to "
                  "defaults from dict".format(name))

    async def module_to_nv(self):
        for value in [not self.save_value, self.save_value, not self.save_value]:
            self.save_signal.set_value(value)
            self.save_frame.update_from_signals()

            result = await self.protocol.write(
                nv_signal=self.save_signal,
                passive=True,
                meta=MetaEnum.value,
            )
            save_signal_response = result[0]
            write_accepted = save_signal_response == value
            if not write_accepted:
                raise Exception('Save to NV write rejected: {}'.format(
                    save_signal_response)
                )

        if self.nv_save_in_progress_signal is not None:
            await self.wait_for_nv_save_complete()

    async def wait_for_nv_save_complete(self):
        while True:
            result = await self.protocol.read(
                nv_signal=self.nv_save_in_progress_signal,
                meta=MetaEnum.value,
            )
            save_active = result[0]
            if not save_active:
                break
            await epyqlib.utils.twisted.sleep(0.5)

    def logger_set_frames(self):
        frames = [frame for frame in self.set_frames.values()
                  if frame.mux_name.startswith('LoggerChunk')]
        frames.sort(key=lambda frame: frame.mux_name)

        return frames

    def signal_from_names(self, frame_name, value_name):
        frame = [f for f in self.set_frames.values()
                 if f.mux_name == frame_name]

        try:
            frame, = frame
        except ValueError as e:
            raise NotFoundError(
                'Frame not found: {}'.format(frame_name)) from e

        signal = [s for s in frame.signals
                   if s.name == value_name]

        try:
            signal, = signal
        except ValueError as e:
            raise NotFoundError(
                'Signal not found: {}:{}'.format(frame_name, value_name)) from e

        return signal

    def cyclic_read_all(self):
        self.cyclic_reader.start()

    def cancel_cyclic_read_all(self):
        self.cyclic_reader.cancel()


@attr.s
class CyclicReader:
    nvs = attr.ib()
    read_call = attr.ib()
    metas = attr.ib()
    pause_requests = attr.ib(factory=weakref.WeakSet)
    _deferred = attr.ib(init=False, default=None)

    def start(self):
        self._deferred = self._cyclic_read_all()

    def cancel(self):
        if self._deferred is not None:
            self._deferred.cancel()
            self._deferred = None

    @contextlib.contextmanager
    def pause_manager(self, id):
        self.pause(id)
        try:
            yield
        finally:
            self.unpause(id)

    def pause(self, id):
        self.pause_requests.add(id)

    def unpause(self, id):
        self.pause_requests.discard(id)

    @epyqlib.utils.twisted.ensure_deferred
    @epyqlib.utils.twisted.errback_dialog
    @epyqlib.utils.twisted.ignore_cancelled
    async def _cyclic_read_all(self):
        x = collections.defaultdict(list)
        for nv in self.nvs:
            x[nv.frame].append(nv)

        while True:
            for meta in self.metas:
                for frame, nvs in x.items():
                    try:
                        while len(self.pause_requests) > 0:
                            await epyqlib.utils.twisted.sleep(0.250)

                        await epyqlib.utils.twisted.sleep(0.02)

                        d, _ = await self.read_call(
                            only_these=nvs,
                            background=True,
                            meta=(meta,),
                        )
                    except (
                            epyqlib.twisted.nvs.CanceledError,
                            epyqlib.twisted.nvs.SendFailedError,
                    ):
                        continue

                    # TODO: CAMPid 0347987975t427567139419439349
                    for nv in nvs:
                        if not nv.status_signal.write_only:
                            value = d[nv.status_signal]
                            nv.set_meta(value, meta=meta, check_range=False)
                            nv.set_from_device(
                                column=getattr(Columns.indexes, meta.name),
                            )


def strip_uuid_from_comment(comment):
    match = re.search(r'<uuid:([a-z0-9-]+)>', comment)

    if match is None:
        return comment, None

    uuid_object = uuid.UUID(match[1])

    stripped_comment = comment.replace(match[0], '').strip()

    return stripped_comment, uuid_object


class Nv(epyqlib.canneo.Signal, TreeNode):
    changed = epyqlib.utils.qt.Signal(
        TreeNode,
        int,
        TreeNode,
        int,
        list,
    )

    def __init__(
        self,
        signal,
        frame,
        parent=None,
        meta=None,
        meta_value=None,
        base=True,
        for_remote_data=True,
    ):
        epyqlib.canneo.Signal.__init__(self, signal=signal, frame=frame,
                                    parent=parent)
        TreeNode.__init__(self)

        if signal.comment is None:
            self.parameter_uuid = None
        else:
            self.comment, self.parameter_uuid = strip_uuid_from_comment(
                self.comment,
            )

        if meta_value is None:
            self.meta_value = MetaEnum.value
        else:
            self.meta_value = meta_value

        default = self.default_value
        if default is None:
            default = 0

        if self.frame is not None:
            self.factory = '<factory>' in (self.comment + self.frame.comment)

        self.reset_value = None

        self.clear(mark_modified=False)

        self.fields = Columns(
            value=self.full_string,
            comment=self.comment,
        )
        if self.frame is not None:
            self.fields.name = '{}:{}'.format(self.frame.mux_name, self.name)

        if meta is None:
            self.meta = Meta()
        else:
            self.meta = meta

        self.base = base
        self.for_remote_data = for_remote_data

        if self.base:
            metas = (
                MetaEnum.user_default,
                MetaEnum.factory_default,
                MetaEnum.minimum,
                MetaEnum.maximum,
            )
            for meta in metas:
                setattr(
                    self.meta,
                    meta.name,
                    Nv(signal, frame=None, meta=self.meta, meta_value=meta, base=False)
                )

            for meta in metas:
                getattr(self.meta, meta.name).set_value(None)

                setattr(
                    self.fields,
                    meta.name,
                    getattr(self.meta, meta.name).full_string,
                )

            self.scratch = Nv(
                signal,
                frame=None,
                meta=self.meta,
                base=False,
                for_remote_data=False,
            )
            self.scratch.set_value(None)
            self.fields.scratch = self.scratch.full_string

        self.write_only = False

        self._stale = self.for_remote_data

        self.stale_role = Qt.ForegroundRole

    def _changed(self, column_start=None, column_end=None, roles=(
            Columns.indexes.value,)):
        if column_start is None:
            column_start = Columns.indexes.value
        if column_end is None:
            column_end = column_start

        self.changed.emit(
            self, column_start,
            self, column_end,
            list(roles),
        )

    def set_from_user(self):
        pass

    def set_from_device(self, column):
        self._set_stale(stale=False, column=column)

    def stale(self, column):
        subnode = self._get_signal_for_column(column)

        if subnode is None:
            return False

        return subnode._stale

    def _set_stale(self, stale, column):
        subnode = self._get_signal_for_column(column)
        was_stale = subnode._stale

        subnode._stale = stale

        if was_stale != subnode._stale:
            self._changed(column, column, roles=[self.stale_role])

    def set_stale(self):
        for column in Columns.indexes:
            subnode = self._get_signal_for_column(column)

            if subnode is not None and subnode.for_remote_data:
                self._set_stale(stale=True, column=column)

    def _get_signal_for_column(self, column):
        if not self.base:
            raise Exception('This should only be called on the base signal')

        if column == Columns.indexes.value:
            return self

        if column == Columns.indexes.scratch:
            return self.scratch

        column_name = Columns().index_from_attribute(column)

        return getattr(self.meta, column_name, None)

    def get_meta_signal(self, meta):
        if meta == MetaEnum.value:
            return self

        return getattr(self.meta, meta.name)

    def get_human_value(self, for_file=False, column=None):
        if column is None:
            column = Columns.indexes.value

        if column == Columns.indexes.scratch:
            signal = self.scratch
        else:
            column_name = Columns().index_from_attribute(column)

            signal = self.get_meta_signal(getattr(MetaEnum, column_name))

        if signal is self:
            return super().get_human_value(for_file=for_file)

        return signal.get_human_value(
            for_file=for_file,
            column=Columns.indexes.value,
        )

    def signal_path(self):
        return self.frame.signal_path() + (self.name,)

    def can_be_saturated(self, meta=MetaEnum.value):
        signal = self.get_meta_signal(meta)

        if signal.value is None:
            return False

        return signal.to_human(signal.value) != signal.saturation_value()

    def saturate(self, meta=MetaEnum.value):
        signal = self.get_meta_signal(meta)

        if not signal.can_be_saturated():
            return

        self.set_meta(signal.saturation_value(), meta=meta, mark_modified=True)

    def saturation_value(self):
        if self.value is None:
            return None

        s = self.to_human(self.value)

        if self.meta.minimum.value is not None:
            s = max(self.to_human(self.meta.minimum.value), s)

        if self.meta.maximum.value is not None:
            s = min(self.to_human(self.meta.maximum.value), s)

        return s

    def can_be_reset(self, meta=MetaEnum.value):
        signal = self.get_meta_signal(meta)
        return signal.reset_value != signal.value

    def reset(self, meta=MetaEnum.value):
        signal = self.get_meta_signal(meta)

        if not signal.can_be_reset():
            return

        self.set_meta(signal.reset_value, meta=meta)

    def check_value(self, value, force=False, check_range=False):
        min_max = {MetaEnum.minimum, MetaEnum.maximum}

        extras = {}

        if self.meta is not None:
            if self.meta.minimum.value is None or self.meta_value in min_max:
                extras['minimum'] = self.to_human(self.raw_minimum)
            else:
                extras['minimum'] = self.meta.minimum.to_human(
                    self.meta.minimum.value,
                )

            if self.meta.maximum.value is None or self.meta_value in min_max:
                extras['maximum'] = self.to_human(self.raw_maximum)
            else:
                extras['maximum'] = self.meta.maximum.to_human(
                    self.meta.maximum.value,
                )

        return super().check_value(
            value=value,
            check_range=check_range,
            **extras,
        )

    def set_value(self, value, force=False, check_range=False):
        ok = self.check_value(
            value=value,
            check_range=check_range,
        )
        if not ok:
            return False

        self.reset_value = value

        min_max = {MetaEnum.minimum, MetaEnum.maximum}

        if self.meta is not None:
            extras = {}
            if self.meta.minimum.value is None or self.meta_value in min_max:
                extras['minimum'] = self.to_human(self.raw_minimum)
            else:
                extras['minimum'] = self.meta.minimum.to_human(
                    self.meta.minimum.value,
                )

            if self.meta.maximum.value is None or self.meta_value in min_max:
                extras['maximum'] = self.to_human(self.raw_maximum)
            else:
                extras['maximum'] = self.meta.maximum.to_human(
                    self.meta.maximum.value,
                )

        super().set_value(
            value=value,
            force=force,
            check_range=check_range,
            **extras,
        )
        self.fields.value = self.full_string
        self._changed()

    def check_data(self, data, check_range=True):
        try:
            if data is None:
                return self.check_value(data)
            else:
                return self.check_human_value(data, check_range=check_range)
        except ValueError:
            return False

        return True

    def set_data(self, data, mark_modified=False, check_range=True):
        if not self.check_data(data=data, check_range=check_range):
            return False

        # self.fields.value = value
        reset_value = self.reset_value
        try:
            if data is None:
                self.set_value(data)
            else:
                self.set_human_value(data, check_range=check_range)
        except ValueError:
            return False
        finally:
            if mark_modified:
                self.reset_value = reset_value
        self.fields.value = self.full_string

        return True

    def check_meta(self, data, meta, *args, **kwargs):
        if meta == MetaEnum.value:
            extras = {}
            if 'check_range' in kwargs:
                extras['check_range'] = kwargs['check_range']
            return self.check_data(data, *args, **extras)

        meta_signal = getattr(self.meta, meta.name)

        return meta_signal.check_data(
            data=data,
            *args,
            **kwargs,
        )

    def set_meta(self, data, meta, *args, **kwargs):
        if data is not None:
            result = self.check_meta(data, meta, *args, **kwargs)
            if not result:
                return result

        if meta == MetaEnum.value:
            return self.set_data(data=data, *args, **kwargs)

        meta_signal = getattr(self.meta, meta.name)

        result = meta_signal.set_data(
            data=data,
            *args,
            **kwargs,
        )
        setattr(self.fields, meta.name, meta_signal.full_string)

        return result

    def can_be_cleared(self, meta=MetaEnum.value):
        signal = self.get_meta_signal(meta)

        return signal.value is not None

    def clear(self, mark_modified=True, meta=MetaEnum.value):
        signal = self.get_meta_signal(meta)

        if not signal.can_be_cleared():
            return

        self.set_meta(None, meta=meta, mark_modified=mark_modified)
        if hasattr(signal, 'status_signal'):
            signal.status_signal.set_value(None)

    def is_factory(self):
        return self.factory

    def is_read_only(self):
        return self.frame.read_write.min > 0 or self.is_summary

    def is_write_only(self):
        return self.frame.read_write.max < 1

    def unique(self):
        # TODO: make it more unique
        return str(self.fields.name) + '__'


class Frame(epyqlib.canneo.Frame, TreeNode):
    _send = epyqlib.utils.qt.Signal(tuple)

    def __init__(self, message=None, tx=False, frame=None,
                 multiplex_value=None, signal_class=Nv, mux_frame=None,
                 parent=None, **kwargs):
        epyqlib.canneo.Frame.__init__(self, frame=frame,
                                   multiplex_value=multiplex_value,
                                   signal_class=signal_class,
                                   set_value_to_default=False,
                                   mux_frame=mux_frame,
                                   parent=parent,
                                   **kwargs)
        TreeNode.__init__(self, parent)

        meta_signals = [
            signal
            for signal in self.signals
            if signal.name == 'Meta'
        ]

        if len(meta_signals) == 0:
            self.meta_signal = None
        else:
            try:
                self.meta_signal, = meta_signals
            except ValueError as e:
                raise TooManyMetaSignalsError(
                    'Too many meta signals in {}:{}'.format(self.name, self.mux_name),
                ) from e

        for signal in self.signals:
            if signal.name in ("ReadParam_command", "ReadParam_status"):
                self.read_write = signal
                break

        for signal in self.signals:
            if signal.name.endswith("_MUX"):
                self.mux = signal
                break

    def signal_path(self):
        if self.mux_name is None:
            return self.name,
        else:
            return self.name, self.mux_name

    def update_from_signals(self, for_read=False, data=None, function=None,
                            only_return=False):
        return super().update_from_signals(
            data=data,
            function=function,
            only_return=only_return,
        )

    def send_now(self, signals):
        self._send.emit(signals)


def get_font_awesome():
    return QtGui.QFont(qtawesome._instance().fontname['fa'])


@attr.s
class Icon:
    character = attr.ib()
    check = attr.ib()
    font = attr.ib(factory=get_font_awesome)


@attr.s
class TransactionAction:
    only_these = attr.ib()
    values = attr.ib()
    callback = attr.ib()
    meta = attr.ib()


class NvModel(epyqlib.pyqabstractitemmodel.PyQAbstractItemModel):
    activity_started = pyqtSignal(str)
    activity_ended = pyqtSignal(str)

    def __init__(self, root, parent=None):
        editable_columns = Columns.fill(False)
        for enumerator in MetaEnum:
            setattr(editable_columns, enumerator.name, True)
        editable_columns.scratch = True

        epyqlib.pyqabstractitemmodel.PyQAbstractItemModel.__init__(
                self, root=root, editable_columns=editable_columns,
                alignment=Qt.AlignVCenter | Qt.AlignLeft, parent=parent)

        self.check_range = True

        self.headers = Columns(name='Name',
                               value='Active',
                               scratch='Scratch',
                               minimum='Min',
                               maximum='Max',
                               user_default='User Default',
                               factory_default='Factory Default',
                               comment='Comment')

        root.activity_started.connect(self.activity_started)
        root.activity_ended.connect(self.activity_ended)

        self.icons = Columns(
            reset=Icon(character='\uf0e2', check='can_be_reset'),
            clear=Icon(character='\uf057', check='can_be_cleared'),
            saturate=Icon(character='\uf066', check='can_be_saturated'),
            factory=Icon(character='\uf084', check='is_factory'),
            read_only=Icon(character='\uf023', check='is_read_only')
        )

        self.meta_columns = {
            getattr(Columns.indexes, enumerator.name)
            for enumerator in MetaEnum
        }

        self.icon_columns = set(
            index
            for index, icon in zip(self.icons.indexes, self.icons)
            if icon is not None
        )

        self.force_action_decorations = False

        self.role_functions[epyqlib.utils.qt.UserRoles.sort] = (
            self.data_sort
        )

        self.transaction_actions = None

        palette = QtGui.QPalette(QtWidgets.QWidget().palette())
        color = QtGui.QColor(palette.windowText().color())
        color.setAlphaF(0.3)
        self.stale_foreground = color

    def all_nv(self):
        return self.root.all_nv()

    def flags(self, index):
        flags = super().flags(index)
        node = self.node_from_index(index)

        if not isinstance(node, epyqlib.nv.Nv) or node.is_read_only():
            flags &= ~Qt.ItemIsEditable

        return flags

    def data_sort(self, index):
        node = self.node_from_index(index)

        return '{}{}'.format(
            'a' if isinstance(node, Group) else 'b',
            self.data_display(index),
        )

    def data_font(self, index):
        icon = self.icons[index.column()]
        if icon is not None:
            return icon.font

        return None

    def data_display(self, index):
        node = self.node_from_index(index)
        column = index.column()
        icon = self.icons[column]
        if icon is not None:
            if self.force_action_decorations:
                return icon.character
            else:
                if isinstance(node, epyqlib.nv.Nv):
                    check = getattr(node, icon.check)
                    if check():
                        return icon.character

        super_result = super().data_display(index)

        return super_result

    def data_foreground(self, index):
        node = self.node_from_index(index)

        if isinstance(node, Nv):
            if node.stale(index.column()):
                return self.stale_foreground

        return None

    def data_tool_tip(self, index):
        if index.column() == Columns.indexes.saturate:
            node = self.node_from_index(index)
            if isinstance(node, epyqlib.nv.Nv):
                if node.can_be_saturated():
                    return node.format_strings(
                        value=node.from_human(node.saturation_value()))[0]
        if index.column() == Columns.indexes.reset:
            node = self.node_from_index(index)
            if isinstance(node, epyqlib.nv.Nv):
                if node.can_be_reset():
                    return node.format_strings(value=node.reset_value)[0]
        elif index.column() == Columns.indexes.comment:
            node = self.node_from_index(index)
            if isinstance(node, epyqlib.nv.Nv):
                comment = node.fields.comment
                if comment is None:
                    comment = ''
                return '\n'.join(textwrap.wrap(comment, 60))

    def dynamic_columns_changed(
            self,
            node,
            columns=None,
            meta=MetaEnum.value,
            roles=(Qt.DisplayRole,),
    ):
        if columns is None:
            columns = dynamic_column_indexes_by_meta[meta]

        for column in columns:
            self.changed(node, column, node, column, roles)

    def saturate_node(self, node, meta=MetaEnum.value):
        node.saturate(meta=meta)
        self.dynamic_columns_changed(node, meta=meta)

    def reset_node(self, node, meta=MetaEnum.value):
        node.reset(meta=meta)
        self.dynamic_columns_changed(node, meta=meta)

    def clear_node(self, node, meta=MetaEnum.value):
        node.clear(meta=meta)
        self.dynamic_columns_changed(node, meta=meta)

    def check_range_changed(self, state):
        self.check_range = state == Qt.Checked

    def start_transaction(self):
        if self.transaction_actions is not None:
            raise SimultaneousTransactionsError(self.transaction_actions)

        self.transaction_actions = collections.defaultdict(list)

    def submit_transaction(self):
        d = twisted.internet.defer.ensureDeferred(self._submit_transaction())
        d.addErrback(epyqlib.utils.twisted.catch_expected)
        d.addErrback(epyqlib.utils.twisted.errbackhook)

    async def _submit_transaction(self):
        try:
            for meta in meta_limits_first:
                actions = self.transaction_actions.get(meta, None)
                if actions is None:
                    continue

                values = {}
                callback = None
                nodes = set()

                for action in actions:
                    values.update(action.values)

                    if action.callback is not None:
                        if callback is None:
                            callback = action.callback
                        elif action.callback != callback:
                            raise Exception('ack')

                    if action.meta is not None:
                        if meta is None:
                            meta = action.meta
                        elif action.meta != meta:
                            raise Exception('ack')

                    nodes = nodes.union(action.only_these)

                if callback is not None:
                    callback = functools.partial(
                        callback,
                        only_these=nodes,
                    )

                await self.root.write_all_to_device(
                    only_these=nodes,
                    values=values,
                    callback=callback,
                    meta=(meta,),
                    # TODO: maybe not really...
                    background=True,
                )
        finally:
            self.transaction_done()

    def cancel_transaction(self):
        self.transaction_actions = None

    def transaction_done(self):
        self.transaction_actions = None

    @contextlib.asynccontextmanager
    async def transaction_manager(self):
        self.start_transaction()
        try:
            yield
        except:
            self.cancel_transaction()
            raise

        await self._submit_transaction()

    def setData(self, index, data, role=None, check_range=None):
        if check_range is None:
            check_range = self.check_range

        column = index.column()

        if data == '':
            data = None

        if column == Columns.indexes.scratch:
            if role == Qt.EditRole:
                node = self.node_from_index(index)
                valid = node.check_data(
                    data,
                    check_range=check_range,
                )
                if valid:
                    node.scratch.set_human_value(data, check_range=check_range)
                    node.fields.scratch = node.scratch.full_string
        elif column in self.meta_columns:
            if role == Qt.EditRole:
                node = self.node_from_index(index)
                meta = getattr(
                    MetaEnum,
                    Columns().index_from_attribute(column),
                )
                valid = node.check_meta(
                    data,
                    meta=meta,
                    check_range=check_range,
                )

                if valid:
                    nodes = (node,)

                    if self.transaction_actions is None:
                        callback = functools.partial(
                            self.update_signals,
                            only_these=nodes,
                        )

                        d = self.root.write_all_to_device(
                            only_these=nodes,
                            values={(node, meta): node.calc_human_value(data)},
                            callback=callback,
                            meta=(meta,),
                        )
                        d.addErrback(epyqlib.utils.twisted.catch_expected)
                        d.addErrback(epyqlib.utils.twisted.errbackhook)
                    else:
                        self.transaction_actions[meta].append(
                            TransactionAction(
                                only_these=nodes,
                                values={(node, meta): node.calc_human_value(data)},
                                callback=self.update_signals,
                                meta=meta,
                            ),
                        )

        return False

    # TODO: CAMPid 0347987975t427567139419439349
    def update_signals(self, arg, only_these):
        d, meta = arg
        model = self

        frame = next(iter(d)).frame

        signals = set(only_these) & set(frame.set_frame.parameter_signals)

        for signal in signals:
            if signal.status_signal in d:
                if not signal.status_signal.write_only:
                    value = d[signal.status_signal]
                    signal.set_meta(value, meta=meta, check_range=False)
                    signal.set_from_device(
                        column=getattr(Columns.indexes, meta.name),
                    )

        for signal in frame.set_frame.parameter_signals:
            QtWidgets.QApplication.instance().processEvents()

            model.dynamic_columns_changed(
                signal,
                columns=(getattr(epyqlib.nv.Columns.indexes, meta.name),)
            )

    @pyqtSlot()
    def module_to_nv(self):
        deferred = twisted.internet.defer.ensureDeferred(self._module_to_nv())
        deferred.addErrback(epyqlib.utils.twisted.errbackhook)

    async def _module_to_nv(self):
        with self.activity_manager(
                start='Requesting save to NV...',
                success='Save to NV succeeded',
                failure='Save to NV failed',
        ):
            await self.root.module_to_nv()

    @pyqtSlot()
    def write_to_module(self):
        # TODO: device or module!?!?
        d = self.root.write_all_to_device()
        d.addErrback(epyqlib.utils.twisted.catch_expected)
        d.addErrback(epyqlib.utils.twisted.errbackhook)

    @pyqtSlot()
    def read_from_module(self):
        d = self.root.read_all_from_device()
        d.addErrback(epyqlib.utils.twisted.catch_expected)
        d.addErrback(epyqlib.utils.twisted.errbackhook)

    @pyqtSlot()
    def write_to_file(self, parent=None):
        filters = [
            ('EPC Parameters', ['epp']),
            ('All Files', ['*'])
        ]
        filename = epyqlib.utils.qt.file_dialog(
            filters, save=True, parent=parent)

        if filename is None:
            return

        if len(filename) > 0:
            with open(filename, 'w') as file:
                d = self.root.to_dict()
                s = json.dumps(d, sort_keys=True, indent=4)
                file.write(s)
                file.write('\n')

                self.activity_ended.emit(
                    'Saved to "{}"'.format(filename)
                )

    @pyqtSlot()
    def write_to_value_set_file(self, parent=None):
        fields = attr.fields(epyqlib.pm.valuesetmodel.ValueSet)
        filters = fields.filters.default
        path = epyqlib.utils.qt.file_dialog(
            filters,
            save=True,
            parent=parent,
        )

        if path is not None:
            value_set = self.root.to_value_set()
            value_set.path = path

            try:
                value_set.save()
            except epyqlib.pm.valuesetmodel.SaveCancelled:
                message = 'Save cancelled'
            else:
                message = 'Saved to "{}"'.format(path)

            self.activity_ended.emit(message)

    @pyqtSlot()
    def read_from_file(self, parent=None):
        filters = [
            ('EPC Parameters', ['epp']),
            ('All Files', ['*'])
        ]
        filename = epyqlib.utils.qt.file_dialog(filters, parent=parent)

        if filename is None:
            return

        if len(filename) > 0:
            with open(filename, 'r') as file:
                s = file.read()
                d = json.loads(s)
                self.root.from_dict(d)

                self.activity_ended.emit(
                    'Loaded from "{}"'.format(filename)
                )

    @contextlib.contextmanager
    def activity_manager(self, start, success, failure):
        self.activity_started.emit(start)

        try:
            yield
        except:
            message = failure
            raise
        else:
            message = success
        finally:
            self.activity_ended.emit(message)

    def read_from_value_set_file(self, parent=None):
        d = twisted.internet.defer.ensureDeferred(
            self._read_from_value_set_file(parent=parent),
        )
        d.addErrback(epyqlib.utils.twisted.catch_expected)
        d.addErrback(epyqlib.utils.twisted.errbackhook)

        return d

    async def _read_from_value_set_file(self, parent=None):
        fields = attr.fields(epyqlib.pm.valuesetmodel.ValueSet)
        filters = fields.filters.default
        # filters = epyqlib.pm.valuesetmodel.ValueSet.filters.default
        path = epyqlib.utils.qt.file_dialog(filters, parent=parent)

        if path is None:
            return

        value_set = epyqlib.pm.valuesetmodel.loadp(path)

        parameter_nodes = value_set.model.root.nodes_by_filter(
            filter=lambda node: isinstance(
                node,
                epyqlib.pm.valuesetmodel.Parameter,
            ),
        )
        only_in_file = {
            (parameter, meta)
            for parameter, meta in itertools.product(
                parameter_nodes,
                epyqlib.nv.MetaEnum,
            )
        }

        d = collections.defaultdict(list)
        for parameter in parameter_nodes:
            d[parameter.name].append(parameter)
            d[parameter.parameter_uuid].append(parameter)

        skip_nodes = (
            self.root.access_level_node,
            self.root.password_node,
        )

        @contextlib.asynccontextmanager
        async def cm():
            with self.activity_manager(
                start='Loading value set from "{}"'.format(path),
                success='Loaded value set from "{}"'.format(path),
                failure='Failed to load value set from "{}"'.format(path),
            ):
                with self.root.cyclic_reader.pause_manager(self):
                    async with self.transaction_manager():
                        yield

        async with cm():
            for child in self.root.all_nv():
                if child in skip_nodes:
                    continue

                name = child.fields.name

                parameters = d.get(child.parameter_uuid, [])
                if len(parameters) == 0:
                    parameters = d.get(name, [])

                not_found_format = (
                    "Nv value named '{}' ({{}}) not found when loading "
                    "from value set".format(name)
                )

                if len(parameters) == 1:
                    parameter, = parameters

                    index = self.index_from_node(child)

                    for meta in MetaEnum:
                        only_in_file.discard((parameter, meta))

                        v = getattr(parameter, meta.name)
                        if v is not None:
                            self.setData(
                                index.siblingAtColumn(
                                    getattr(Columns.indexes, meta.name),
                                ),
                                v,
                                Qt.EditRole,
                                check_range=False,
                            )
                        else:
                            logger.warning(not_found_format.format(meta.name))
                elif len(parameters) > 1:
                    logger.warning(
                        "Nv value named '{}' occurred {} times when loading "
                        "from value set".format(name, len(parameters)),
                    )
                else:
                    logger.warning(
                        "Nv value named '{}' not found when loading from "
                        "value set".format(name),
                    )

            for parameter, meta in sorted(only_in_file):
                logger.warning("Unrecognized NV value named '{}' ({}) found when loading "
                      "from value set".format(parameter.name, meta.name))


if __name__ == '__main__':
    import sys

    logger.info('No script functionality here')
    sys.exit(1)     # non-zero is a failure



import io
import textwrap

from PyQt5 import QtWidgets
import PyQt5.uic

import epyqlib.filesview


def test_init(qtbot):
    view = epyqlib.filesview.FilesView()
    qtbot.add_widget(view)

    view.show()
    qtbot.wait_for_window_shown(widget=view)

    assert view.parent() is None
    assert isinstance(view.ui, epyqlib.filesview.Ui)
    assert not hasattr(view.ui, 'something')


def test_build(qtbot):
    view = epyqlib.filesview.FilesView.build()
    qtbot.add_widget(view)

    view.show()
    qtbot.wait_for_window_shown(widget=view)

    assert view.parent() is None
    assert isinstance(view.ui, epyqlib.filesview.Ui)
    assert hasattr(view.ui, 'something')


def test_qt_build(qtbot):
    widget = QtWidgets.QWidget()
    qtbot.add_widget(widget)

    view = epyqlib.filesview.FilesView.qt_build(parent=widget)
    qtbot.add_widget(view)

    view.show()
    qtbot.wait_for_window_shown(widget=view)

    assert view.parent() is widget
    assert isinstance(view.ui, epyqlib.filesview.Ui)
    assert hasattr(view.ui, 'something')


def test_from_ui(qtbot):
    ui_src = io.StringIO(textwrap.dedent("""\
    <?xml version="1.0" encoding="UTF-8"?>
    <ui version="4.0">
     <class>Form</class>
     <widget class="QWidget" name="Form">
      <property name="geometry">
       <rect>
        <x>0</x>
        <y>0</y>
        <width>400</width>
        <height>300</height>
       </rect>
      </property>
      <property name="windowTitle">
       <string>Form</string>
      </property>
      <layout class="QGridLayout" name="gridLayout">
       <item row="0" column="0">
        <widget class="FilesViewQtBuilder" name="widget" native="true"/>
       </item>
      </layout>
     </widget>
     <customwidgets>
      <customwidget>
       <class>FilesViewQtBuilder</class>
       <extends>QWidget</extends>
       <header>epyqlib.filesview</header>
      </customwidget>
     </customwidgets>
     <resources/>
     <connections/>
    </ui>
    """))

    widget = PyQt5.uic.loadUi(ui_src)
    qtbot.add_widget(widget)

    view = widget.widget

    view.show()
    qtbot.wait_for_window_shown(widget=view)

    assert view.parent() is widget
    assert isinstance(view.ui, epyqlib.filesview.Ui)
    assert hasattr(view.ui, 'something')



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'txrx.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1500, 600)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.tx = TxRxView(self.centralwidget)
        self.tx.setObjectName("tx")
        self.verticalLayout.addWidget(self.tx)
        self.rx = TxRxView(self.centralwidget)
        self.rx.setObjectName("rx")
        self.verticalLayout.addWidget(self.rx)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1500, 39))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
from epyqlib.txrxview import TxRxView



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

from PyQt5.QtCore import pyqtProperty
from PyQt5.QtWidgets import QWidget
from PyQt5.QtGui import QColor

import epyqlib.compoundscale_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class CompoundScale(QWidget):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        # If true, flips the compound scale on y axis if vertically oriented.
        self.cs_vertically_flipped = False

        self.ui = epyqlib.compoundscale_ui.Ui_Form()
        self.ui.setupUi(self)

        self.ui.command.in_designer = in_designer
        self.ui.echo.in_designer = in_designer
        self.ui.status.in_designer = in_designer
        self.ui.numeric_status.in_designer = in_designer

        self.update_echo_visibility()

    def update_echo_visibility(self):
        hidden = len(self.echo_signal_path) == 0
        self.ui.echo.setHidden(hidden)
        self.ui.echo.ignore = True

    @pyqtProperty('QString')
    def command_signal_path(self):
        return self.ui.command.signal_path

    @command_signal_path.setter
    def command_signal_path(self, value):
        self.ui.command.signal_path = value

    @pyqtProperty('QString')
    def echo_signal_path(self):
        return self.ui.echo.signal_path

    @echo_signal_path.setter
    def echo_signal_path(self, value):
        self.ui.echo.signal_path = value
        self.update_echo_visibility()

    @pyqtProperty('QString')
    def status_signal_path(self):
        return self.ui.status.signal_path

    @status_signal_path.setter
    def status_signal_path(self, value):
        self.ui.status.signal_path = value
        self.ui.numeric_status.signal_path = value

    @pyqtProperty(bool)
    def status_override_range(self):
        return self.ui.status.override_range

    @status_override_range.setter
    def status_override_range(self, override):
        self.ui.status.override_range = override

    @pyqtProperty(float)
    def status_minimum(self):
        return self.ui.status.minimum

    @status_minimum.setter
    def status_minimum(self, min):
        self.ui.status.minimum = float(min)

    @pyqtProperty(float)
    def status_maximum(self):
        return self.ui.status.maximum

    @status_maximum.setter
    def status_maximum(self, max):
        self.ui.status.maximum = float(max)

    @pyqtProperty(str)
    def status_label(self):
        return self.ui.numeric_status.label_override

    @status_label.setter
    def status_label(self, label):
        self.ui.numeric_status.label_override = label

    # cs_flipped function allows for the qscale to be flipped or not when
    # in vertical orientation
    @pyqtProperty(bool)
    def cs_flipped(self):
        return self.cs_vertically_flipped

    @cs_flipped.setter
    def cs_flipped(self, value):
        self.cs_vertically_flipped = value
        self.ui.status.scale.vertically_flipped = value

    @pyqtProperty(float)
    def lower_red_breakpoint(self):
        return self.ui.status._breakpoints[0]

    @lower_red_breakpoint.setter
    def lower_red_breakpoint(self, breakpoint):
        self.ui.status._breakpoints[0] = breakpoint
        self.ui.status.update_configuration()

    @pyqtProperty(float)
    def lower_yellow_breakpoint(self):
        return self.ui.status._breakpoints[1]

    @lower_yellow_breakpoint.setter
    def lower_yellow_breakpoint(self, breakpoint):
        self.ui.status._breakpoints[1] = breakpoint
        self.ui.status.update_configuration()

    @pyqtProperty(float)
    def upper_yellow_breakpoint(self):
        return self.ui.status._breakpoints[2]

    @upper_yellow_breakpoint.setter
    def upper_yellow_breakpoint(self, breakpoint):
        self.ui.status._breakpoints[2] = breakpoint
        self.ui.status.update_configuration()

    @pyqtProperty(float)
    def upper_red_breakpoint(self):
        return self.ui.status._breakpoints[3]

    @upper_red_breakpoint.setter
    def upper_red_breakpoint(self, breakpoint):
        self.ui.status._breakpoints[3] = breakpoint
        self.ui.status.update_configuration()

    @pyqtProperty(QColor)
    def lower_red_color(self):
        return self.ui.status._colors[0]

    @lower_red_color.setter
    def lower_red_color(self, color):
        self.ui.status._colors[0] = color
        self.ui.status.update_configuration()

    @pyqtProperty(QColor)
    def lower_yellow_color(self):
        return self.ui.status._colors[1]

    @lower_yellow_color.setter
    def lower_yellow_color(self, color):
        self.ui.status._colors[1] = color
        self.ui.status.update_configuration()

    @pyqtProperty(QColor)
    def green_color(self):
        return self.ui.status._colors[2]

    @green_color.setter
    def green_color(self, color):
        self.ui.status._colors[2] = color
        self.ui.status.update_configuration()

    @pyqtProperty(QColor)
    def upper_yellow_color(self):
        return self.ui.status._colors[3]

    @upper_yellow_color.setter
    def upper_yellow_color(self, color):
        self.ui.status._colors[3] = color
        self.ui.status.update_configuration()

    @pyqtProperty(QColor)
    def upper_red_color(self):
        return self.ui.status._colors[4]

    @upper_red_color.setter
    def upper_red_color(self, color):
        self.ui.status._colors[4] = color
        self.ui.status.update_configuration()

if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import os
import re

from PyQt5 import QtCore
from PyQt5.QtCore import pyqtProperty, QFile, QFileInfo, QEvent
from PyQt5.QtGui import QColor
from PyQt5.QtXml import QDomDocument

import epyqlib.widgets.abstractwidget
import epyqlib.widgets.led_ui


# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


darker_factor = 400


class FontChangeEventSignal(QtCore.QObject):
    triggered = QtCore.pyqtSignal()

    def eventFilter(self, _, event):
        if event.type() == QtCore.QEvent.FontChange:
            self.triggered.emit()

        return False


def set_attribute_recursive(element, tag_name, attribute,
                            new_color):
    if element.tagName() == tag_name:
        text = element.attribute(attribute)
        text = re.sub('(?<=fill:#)[0-9A-F]{6}', new_color, text)
        element.setAttribute(attribute, text)

    child_nodes = element.childNodes()
    children = [child_nodes.at(i) for i in range(child_nodes.length())]
    for child in children:
        if child.isElement():
            set_attribute_recursive(element=child.toElement(),
                                    tag_name=tag_name,
                                    attribute=attribute,
                                    new_color=new_color)


def make_color(svg_string, new_color):
    doc = QDomDocument()
    doc.setContent(svg_string)
    set_attribute_recursive(element=doc.documentElement(),
                            tag_name="circle",
                            attribute="style",
                            new_color=new_color)

    return doc.toByteArray()


def rgb_string(color):
    return ('{:02X}' * 3).format(*color.getRgb())


class Led(epyqlib.widgets.abstractwidget.AbstractWidget):
    def __init__(self, parent=None, in_designer=False):
        file_name = 'led.svg'

        # TODO: CAMPid 9549757292917394095482739548437597676742
        if not QFileInfo(file_name).isAbsolute():
            file = os.path.join(
                QFileInfo.absolutePath(QFileInfo(__file__)), file_name)
        else:
            file = file_name
        file = QFile(file)
        file.open(QFile.ReadOnly | QFile.Text)
        self.svg_string = file.readAll()

        # TODO: shouldn't this be in AbstractWidget?
        self._on_value = 1
        self._relative_height = 1
        self._label_from_enumeration = False

        self._value = False
        self.svg = {
            'on': None,
            'automatic_off': None,
            'manual_off': None
        }

        self._last_loaded_svg = object()

        self._on_color = QColor()
        self._manual_off_color = QColor()
        self._automatic_off_color = True

        super().__init__(
            ui_class=epyqlib.widgets.led_ui.Ui_Form,
            parent=parent,
            in_designer=in_designer,
        )

        self.ui.value.main_element = 'led'
        self.update_svg()

        self.on_color = QColor("#20C020")
        self.manual_off_color = self.on_color.darker(factor=darker_factor)

        self.font_event_signal = FontChangeEventSignal()
        self.ui.label.installEventFilter(self.font_event_signal)
        self.font_event_signal.triggered.connect(self.update_svg)

    @pyqtProperty(bool)
    def label_from_enumeration(self):
        return self._label_from_enumeration

    @label_from_enumeration.setter
    def label_from_enumeration(self, from_enumeration):
        self._label_from_enumeration =  bool(from_enumeration)

        # TODO: this is a hacky way to trigger an update
        self.set_signal(signal=self.signal_object,
                        force_update=True)

    @pyqtProperty(int)
    def on_value(self):
        return self._on_value

    @on_value.setter
    def on_value(self, new_on_value):
        new_on_value = int(new_on_value)
        if self._on_value != new_on_value:
            self._on_value = new_on_value
            self.update_svg()

            # TODO: this is a hacky way to trigger an update
            self.set_signal(signal=self.signal_object,
                            force_update=True)

    @pyqtProperty(bool)
    def automatic_off_color(self):
        return self._automatic_off_color

    @automatic_off_color.setter
    def automatic_off_color(self, automatic):
        automatic = bool(automatic)
        if automatic != self._automatic_off_color:
            self._automatic_off_color = bool(automatic)
            self.update_svg()

    @pyqtProperty(QColor)
    def on_color(self):
        return self._on_color

    @on_color.setter
    def on_color(self, new_color):
        self._on_color = QColor(new_color)

        self.svg['on'] = make_color(self.svg_string, rgb_string(self._on_color))

        if self.automatic_off_color:
            self.svg['automatic_off'] = make_color(
                self.svg_string,
                rgb_string(self._on_color.darker(factor=darker_factor))
            )

        self.update_svg()

    @pyqtProperty(QColor)
    def manual_off_color(self):
        return self._manual_off_color

    @manual_off_color.setter
    def manual_off_color(self, new_color):
        self._manual_off_color = QColor(new_color)

        self.svg['manual_off'] = make_color(self.svg_string, rgb_string(self._manual_off_color))

        self.update_svg()

    @pyqtProperty(float)
    def relative_height(self):
        return self._relative_height

    @relative_height.setter
    def relative_height(self, multiplier):
        self._relative_height = multiplier

        self.update_svg()

    def set_value(self, value):
        # TODO: quit hardcoding this and it's better implemented elsewhere
        if self.signal_object is not None:
            value = self.signal_object.value
        else:
            value = value

        self._value = (value == self.on_value)

        self.update_svg()

    def update_svg(self):
        if self._value:
            svg = self.svg['on']
        elif self.automatic_off_color:
            svg = self.svg['automatic_off']
        else:
            svg = self.svg['manual_off']

        if svg is not self._last_loaded_svg:
            if svg is None:
                self.ui.value.load(b'')
            else:
                self.ui.value.load(svg)
            self._last_loaded_svg = svg

        # TODO: figure out if this should be needed.  seems not but
        #       without it the font metric can return invalid values
        #       at least when the label_visible property is not
        #       explicitly set in the .ui file and the font size
        #       has been set via a second stylesheet.
        #
        #       Possibly:
        #           https://bugreports.qt.io/browse/QTBUG-42720

        self.ui.label.setVisible(self.ui.label.isVisibleTo(self))
        height = self.relative_height * self.ui.label.fontMetrics().height()

        ratio = self.ui.value.ratio()
        if ratio != 0:
            width = height / ratio
        else:
            width = 0

        self.ui.value.setFixedSize(width, height)

    def set_label_custom(self, new_signal=None):
        label = None

        if new_signal is not None:
            if self.label_from_enumeration and new_signal is not None:
                label = new_signal.enumeration[self.on_value]

        return label

    def event(self, *args, **kwargs):
        result = epyqlib.widgets.abstractwidget.AbstractWidget.event(
            self, *args, **kwargs
        )

        event = args[0]
        if event.type() == QEvent.Polish:
            self.update_svg()

        return result

    def set_unit_text(self, units):
        pass


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure
